#!/usr/bin/python3

import sys
import scipy.io
import json
import os
import numpy as np
import tensorflow as tf
from universal_tf_rnn.aux_code.ops import pad_and_mask
from universal_tf_rnn.aux_code.tf_ops import get_blank_state
from universal_tf_rnn.model_file import RNN_Model

# Remove warnings about in code not being compiled with some optimizations
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
tf.logging.set_verbosity(tf.logging.ERROR)

record = sys.argv[1]

# Read waveform samples (input is in WFDB-MAT format)
mat_data = scipy.io.loadmat(record + ".mat")
x = mat_data['val'][0].astype(np.float32)

# Load the config file
with open('model/config.json', 'r') as fp:
    args = json.load(fp)

# Prepare data
# Normalize
mean = np.average(x)
x = (x - mean) / np.std(x - mean)
# mean = 7.53577291423
# std = 237.287476288
# x = (x - mean) / std
x = np.expand_dims(x, -1)


if args['truncate_size'] is None:
    print('Segmenting signal ... ')
    min_len = 400

    x_new = []
    for i in range(0, len(x), args['max_sequence_length']):
        if len(x[i:i + args['max_sequence_length']]) > min_len:
            x_new.append(x[i:i + args['max_sequence_length']])

    x, _, sequence_lengths = pad_and_mask(x_new)
else:
    # If this was trained using TBPTT, classify the entire signal
    print('Analysing entire signal ... ')
    x = np.expand_dims(x, 0)
    sequence_lengths = np.array([len(x)])

# Start session and load model with config
h_dim = list(map(int, args['layers'].split(',')))
model = RNN_Model(x.shape[-1], n_classes=4, h_dim=h_dim,
                  max_sequence_length=args['max_sequence_length'],
                  is_test=True,
                  batch_norm=args['batch_norm'])

states = get_blank_state(x.shape[0], hidden_dim=h_dim)

input_feed = {model.x: x,
              model.sequence_lengths: sequence_lengths,
              model.training: False,
              model.keep_prob: 1.0,
              model.state_c: tuple(s.c for s in states),
              model.state_h: tuple(s.h for s in states)
              }

with tf.Session() as sess:
    saver = tf.train.Saver()  # Imperative to only create a saver
    # and not import a meta_graph
    # saver.restore replaces init()
    saver.restore(sess, tf.train.latest_checkpoint('model/'))

    output_probs = sess.run(model.output_probs, input_feed)
    # output_probs has shape (batch_size, n_classes)

# Classify
labels = ["N", "A", "O", "~"]
# Get the average classifications over the batch
output_probs = np.average(output_probs, axis=0)

answer = labels[np.argmax(output_probs)]


# Write result to answers.txt
answers_file = open("answers.txt", "a")
answers_file.write("%s,%s\n" % (record, answer))
answers_file.close()
