import numpy as np
import random
import copy
import json
import sys
import time
import itertools
try:
    from sklearn import metrics
except ImportError:
    print("Could not import sklearn.metrics")

seed = 42
np.random.seed(seed)


def timeit(f):
    '''
    A function to time the execution of a function and print the elapsed time.
    Example usage:

        @timeit
        def function_to_time(normal_args):
            do foo
            return bar
    '''
    def timed(*args, **kwargs):
        start_time = time.time()
        result = f(*args, **kwargs)
        end_time = time.time()

        print("   [-] %s : %2.5f sec" % (f.__name__, end_time - start_time))
        return result

    return timed


def get_class_balance(y):
    '''
    Calculate the percentages of each class present in the data.

    Input:
    y - A numpy array of the labels, could be either one_hot encoded or
        cardinal.

    Returns:
    A numpy array of percentages in each class.
    '''
    if len(y.shape) == 1:
        y = one_hot_encoder(y)
    return np.average(y, axis=0)


def save_config_dict(config, path, tags=None, values=None):
    '''
    Save a dictionary (config) to a json file in path

    Input:
    config  - A dictionary to save
    path    - String specifying the path
    tags    - (optional) The keys to add to the dict
    values  - (optional) The values to add to the dict
    '''
    if '.json' not in path:
        path = path+'config.json'

    if tags is not None and values is not None:
        for tag, val in zip(tags, values):
            if type(val) is np.ndarray:
                config[tag] = list(val)
            else:
                config[tag] = val

    with open(path, 'w') as fp:
        json.dump(config, fp, indent=4, sort_keys=True)


def create_conf_dict(args):
    '''
    Create the config dictionary from tensorflow flags.
    '''
    d = copy.deepcopy(vars(args))  # Important to copy the dict,
                                   # otherwise the original args
                                   # dictionary will be mutated.
    del d['logdir']
    del d['test']
    del d['gpu']

    return d


def create_conf_dict_from_flags(FLAGS):
    '''
    Create the config dictionary from tensorflow flags.
    '''
    d = copy.deepcopy(FLAGS.__flags)  # Important to copy the flags dict,
                                      # otherwise the original FLAGS
                                      # dictionary will be mutated.
    del d['logdir']
    del d['test']
    del d['gpu']

    return d


def calculate_metrics(y_true, y_pred, labels=None):
    '''
    A function to compute the accuracy, F1 score, and AUC.
    The 'macro' F1 score is computed; meaning the F1 score is computed for each
    class and then the unwieghted average is returned.

    Input:
    y_true - a numpy array of true labels.
    y_pred - a numpy array of predicted labels.
    labels - the labels for each class in y_true with shape (max(y_true)+1).

    Returns:
    A numpy array of Accuracy, F1, AUC.
    '''
    if len(y_true.shape) == 2:
        y_true = np.argmax(y_true, axis=1)

    if len(y_pred.shape) == 2:
        y_pred = np.argmax(y_pred, axis=1)

    acc = np.average(np.equal(y_true, y_pred))

    F1 = metrics.f1_score(y_true, y_pred, labels, average='macro')

    #  AUC = metrics.roc_auc_score(one_hot_encoder(y_true),
                                #  one_hot_encoder(y_pred))
    AUC = 0

    return np.array([acc, F1, AUC])


def get_max_length(var):
    '''
    A function that returns the largest length of an array in a list of arrays.

    Input:
    var - A list of numpy arrays

    Returns:
    The largest length (int)
    '''
    return len(max(var, key=lambda v: len(v)))


def print_metrics(tags, values):
    '''
    Prints the tags and values column-wise
    The columns have a width of 13 characters and tags should not be
    longer than 10 characters
    '''
    label_str = ""
    value_str = ''
    for i in range(len(tags)):
        label_str += tags[i].rjust(13)
        value_str += '{:13.4}'.format(values[i])

    print(label_str)
    print(value_str)


class BatchTruncator(object):
    '''
    A generator wrapper for Truncate BackPropagation Through Time.
    '''
    def __init__(self, x, seq_lengths, truncate_size):
        '''
        Inputs:
        x -- The numpy array of data to truncate, shape
        (batch_size, sequence_length, n_features)
        seq_lengths -- Numpy array of sequence lengths of samples in x
                       with shape (batch_size)
        truncate_size -- The number of steps after which the batch is
        truncated
        '''
        self.iterator = data_iterator(x, seq_lengths, truncate_size)

    def __next__(self):
        return next(self.iterator)


def data_iterator(x, seq_lengths, truncate_size):
    '''
    Inputs:
    x -- The numpy array of data to truncate, shape
    (batch_size, sequence_length, n_features)
    seq_lengths -- Numpy array of sequence lengths of samples in x
                   with shape (batch_size)
    truncate_size -- The number of steps after which the batch is
    truncated
    '''
    num_truncations = x.shape[1] // truncate_size
    # Note: The remaining section of the signal is ignored

    if num_truncations == 0:
        raise ValueError('Number of truncations == 0, decrease the'
                         ' truncate_size.')

    for i in range(num_truncations):
        truncated_x = x[:, i*truncate_size: (i + 1) * truncate_size]
        seq_lengths = np.maximum(seq_lengths-truncate_size,
                                 np.zeros_like(seq_lengths))
        new_lengths = np.minimum(seq_lengths,
                                 np.ones_like(seq_lengths)*truncate_size)
        yield truncated_x, new_lengths


def get_minibatches_indices(n, minibatch_size, shuffle=True,
                            shuffleBatches=True,
                            allow_smaller_final_batch=False):
    # shuffleBatches: when true, this shuffles the order of the batches. So
    # every iteration sees a different batch order, with different sequence
    # orders.
    """
    Used to shuffle the dataset at each iteration.

    Inputs:
    n - An integer corresponding to the total number of samples
    minibatch_siez      - The size of the minibatches to create.
    allow_smaller_final_batch
                        - [boolean] if true, the fina
                        an entire batch will be returned as the final minibatch

    Returns:
    minibatches - a list of lists with indices in the lowest dimension
    """
    idx_list = np.arange(n, dtype="int32")

    if shuffle:
        np.random.shuffle(idx_list)

    minibatches = []
    minibatch_start = 0

    for i in range(n // minibatch_size):
        minibatches.append(idx_list[minibatch_start:
                                    minibatch_start + minibatch_size])
        minibatch_start += minibatch_size
    if allow_smaller_final_batch:
        if (minibatch_start != n):
            # Make a minibatch out of what is left
            minibatches.append(idx_list[minibatch_start:])

    if shuffleBatches is True:
        # minibatches here is a list of list indexes
        random.shuffle(minibatches)

    return minibatches


def random_split(n, test_frac=0.1):
    all_idx = np.arange(n)
    test_idx = all_idx[np.random.choice(
        n, int(np.ceil(test_frac * n)), replace=False)]
    train_idx = np.setdiff1d(all_idx, test_idx)
    assert (np.all(np.sort(np.hstack([train_idx, test_idx])) == all_idx))
    return train_idx, test_idx


def onehot_categorical(batchsize, n_labels):
    y = np.zeros((batchsize, n_labels), dtype=np.float32)
    indices = np.random.randint(0, n_labels, batchsize)
    for b in range(batchsize):
        y[b, indices[b]] = 1
    return y


def randomly_split_data(Y, test_frac=0.5, valid_frac=0, verbose=0, equalClasses=False, undersample_training_set=False):
    '''
    Split the data into 3 sets:
        test_frac and valid_frac are the percentage of the original dataset that is reserved for test and validation datasets.

    Inputs:
    Y           - The one-hot encoded labels of the data with shape [n_samples, n_classes]

    Returns:
    A tuple with
        train-idx       - A list of the train indices.
        valid-idx       - A list of the validation indices.
        test-idx        - A list of the testing indices.
    '''
    split = None
    # Used to determine the smallest number of allowable samples in the splits
    smallest_class = min(Y.sum(axis=0))

    if verbose > 0:
        sys.stdout.write('Generating {0} test split'.format(
            test_frac + valid_frac))
        sys.stdout.flush()
    while split is None:
        if verbose > 0:
            sys.stdout.write('.')
            sys.stdout.flush()
        not_test_idx, test_idx = random_split(
            Y.shape[0], test_frac=test_frac + valid_frac)

        # The oversampling does not result in a perfect 50:10:40 split
        if equalClasses == True:  # This process is called oversampling. It is used to create even class spread in the training set
            # Find the largest represented class in training set
            largestIndex = np.argmax(Y[not_test_idx, :].sum(axis=0))
            largest = max(Y[not_test_idx, :].sum(axis=0))
            print('testing shapes:', Y[test_idx, :].sum(axis=0))
            print('training sets:', Y[not_test_idx, :].sum(axis=0))
            for i in range(Y.shape[1]):  # Loop over all the other classes
                if i != largestIndex:
                    # Don't let the corresponding test set of the class become
                    # smaller than 0.2 of the smallest class
                    while Y[not_test_idx, i].sum(axis=0) < largest and Y[test_idx, i].sum(axis=0) > 0.2 * smallest_class:
                        temp_idx = test_idx[Y[test_idx, i] == 1]
                        #diff = largest - Y[not_test_idx, i].sum(axis=0)
                        # print(temp_idx[0])
                        not_test_idx = np.append(not_test_idx, temp_idx[0])
                        #not_test_idx = np.append(not_test_idx, temp_idx[0:diff])
                        #test_idx = test_idx[test_idx != temp_idx[0:diff]]
                        test_idx = test_idx[test_idx != temp_idx[0]]
                        #print(test_idx != temp_idx[0])
                    print('testing shapes:', Y[test_idx, :].sum(axis=0))
                    print('training sets:', Y[not_test_idx, :].sum(axis=0))
            # Here we undersample the testing set. This makes the number of samples in each class for the testing set equal by removing the excess in classes with more than the minimum
            # I could also instead of merely removing the indices, add them to
            # the validation set
            # Find the largest represented class in training set
            smallestIndex = np.argmin(Y[test_idx, :].sum(axis=0))
            smallest = min(Y[test_idx, :].sum(axis=0))
            collected_idx = []
            for i in range(Y.shape[1]):  # Loop over all the other classes
                if i != smallestIndex:
                    while Y[test_idx, i].sum(axis=0) > smallest:
                        temp_idx = test_idx[Y[test_idx, i] == 1]
                        # not_test_idx = np.append(not_test_idx, temp_idx[0])
                        # #could replace with validation set
                        test_idx = test_idx[test_idx != temp_idx[0]]
                    print('testing shapes:', Y[test_idx, :].sum(axis=0))
                    print('training sets:', Y[not_test_idx, :].sum(axis=0))

        if undersample_training_set == True:
            # Find the largest represented class in training set
            smallest_index = np.argmin(Y[not_test_idx, :].sum(axis=0))
            smallest = min(Y[not_test_idx, :].sum(axis=0))
            print('testing shapes:', Y[test_idx, :].sum(axis=0))
            print('training sets:', Y[not_test_idx, :].sum(axis=0))
            for i in range(Y.shape[1]):  # Loop over all the other classes
                if i != smallest_index:
                    while Y[not_test_idx, i].sum(axis=0) > smallest:
                        #print(Y[not_test_idx, i].sum(axis=0))
                        temp_idx = not_test_idx[Y[not_test_idx, i] == 1]
                        # Add the sample to the test index
                        test_idx = np.append(test_idx, temp_idx[0])
                        # Remove the sample from the training index
                        not_test_idx = not_test_idx[
                            not_test_idx != temp_idx[0]]
                    print('testing shapes:', Y[test_idx, :].sum(axis=0))
                    print('training sets:', Y[not_test_idx, :].sum(axis=0))

        # assert(np.all(np.sort(np.hstack([not_test_idx,test_idx])) == np.arange(Y.shape[0]))) #makes sure that all the indexes are there
        # Above is not relevant when I cull the datasets to have equal classes.
        # makes sure that there are more than x*smallest_class exapmles of each
        # label in each set.
        if np.all(Y[not_test_idx, :].sum(axis=0) >= 0.8 * (1 - test_frac - valid_frac) * smallest_class) and np.all(Y[test_idx, :].sum(axis=0) >= 0.8 * (test_frac + valid_frac) * smallest_class):
            if valid_frac != 0:
                if verbose > 0:
                    sys.stdout.write(
                        'Generating {0}/{1} train/test splits'.format(1 - (test_frac + valid_frac), valid_frac))
                    sys.stdout.flush()
                while split is None:
                    if verbose > 0:
                        sys.stdout.write('.')
                        sys.stdout.flush()
                    final_test_idx, valid_idx = random_split(
                        Y[test_idx].shape[0], test_frac=valid_frac / (test_frac + valid_frac))
                    assert(np.all(np.sort(np.hstack((final_test_idx, valid_idx))) == np.arange(
                        Y[test_idx].shape[0])))
                    if np.all(Y[test_idx, :][final_test_idx, :].sum(axis=0) >= 0.6 * test_frac * smallest_class) and np.all(Y[test_idx, :][valid_idx, :].sum(axis=0) >= 0.6 * valid_frac * smallest_class):
                        split = (np.sort(not_test_idx), np.sort(
                            test_idx[valid_idx]), np.sort(test_idx[final_test_idx]))
                        print('Split completed.\n')
                        break
                    else:
                        print(
                            'Labels were not split evenly for the validation set, resplitting...\n')
                        print(Y[test_idx, :][final_test_idx, :].sum(axis=0))
                        print(0.6 * test_frac * smallest_class)
                        print(Y[test_idx, :][valid_idx, :].sum(axis=0))
                        print(0.6 * valid_frac * smallest_class)
            else:
                split = (np.sort(not_test_idx), None, np.sort(test_idx))
                print('Split completed.\n')
        else:
            print('Labels were not split evenly for the test set, resplitting...\n')
            print(Y[not_test_idx, :].sum(axis=0))
            print(0.7 * (1 - test_frac) * smallest_class)
            print(Y[test_idx, :].sum(axis=0))
            print(0.7 * test_frac * smallest_class)

    # Sanity check
    if valid_frac != 0:
        print("Train:Validation:Testing - %d:%d:%d" % (len(split[0]), len(split[1]),
                                                       len(split[2])))
    else:
        print("Train:Testing - %d:%d" % (len(split[0]), len(split[2])))

    return split


def one_hot_encoder(labels, n_classes=None):
    '''
    A function that converts an array of integers, representing the labels
    of data, into an array of one-hot encoded vectors.

    Inputs:
    Labels      - a numpy array of integers (nsamples)
    n_classes   - (int) the number of classes if labels does not contain all

    Returns:
    A numpy array of one-hot encoded vectors with size [nsamples, n_classes]
    '''
    if n_classes is None:
        n_classes = int(labels.max() + 1)

    one_hot = np.zeros((labels.shape[0], n_classes))
    one_hot[np.arange(labels.shape[0]), labels.astype(int)] = 1
    return one_hot


class DataDistribution(object):

    def __init__(self):
        self.mu = 0
        self.sigma = 5.0

    def sample(self, N):
        samples = np.random.normal(self.mu, self.sigma, N)
        samples.sort()
        return samples


class CategoricalDistribution(object):
    # Draw samples from a categorical distribution
    # The sample will be a vector of probabilities summing to 1.

    # def __init__(self):
    # Set the number samples to draw to determine the categorical probabilities.
    # self.number_of_draws = 1000

    def sample(self, batch_size, N):
        '''
        The categorical samples are drawn with equal probability at random.

        Inputs:
        N           - The number of classes (integer)
        batch_size  - The numbe of samples in the batch (int)

        Returns:
        samples     - A vector of probabilites with shape [batch_size, N]
        '''
        samples = np.random.multinomial(
            3 * N, [1 / float(N)] * N, size=batch_size)
        # The total of each row in samples will be 3N
        # We divide by 3N to make the total equal 1
        samples = samples / float(3 * N)

        # ## One-hot version
        # samples = np.zeros((batch_size,N))

        # # For each sample set a ranom entry to 1
        # for i in range(batch_size):
        #     samples[i,np.random.randint(0,N)] = 1

        return samples


def pad_and_mask(seqs, labels=None, maxlen=None):
    """Create the matrices from the datasets.

    This pad each sequence to the same length: the length of the
    longest sequence or maxlen.

    if maxlen is set, we will cut all sequence to this maximum
    length.

    Input:
    labels - a list of list (nsamples x ndim)
    seqs - list of sequences

    Output:
    x -- a numpy array with shape (batch_size, max_time_steps, num_features)
    x_mask -- a numpy array with shape (batch_size, max_time_steps) where
                unit entries correspond to elements present in the corresponding
                sequence
    Also returns an array of the sequence lengths

    """
    lengths = [len(s) for s in seqs]

    if maxlen is not None:  # Creates a shortened matrix with all sequences shorter or as long as maxlen
        new_lengths = lengths
        for i in range(len(lengths)):
            if lengths[i] > maxlen:
                seqs[i] = seqs[i][:maxlen]
                new_lengths[i] = maxlen
        lengths = new_lengths

        if len(lengths) < 1:
            return None, None, None

    # print(lengths)
    n_samples = len(seqs)
    maxlen = np.max(lengths)
    inputDimSize = seqs[0].shape[1]

    x = np.zeros((n_samples, maxlen, inputDimSize))
    x_mask = np.zeros((n_samples, maxlen))

    for idx, s in enumerate(seqs):
        # y[:lengths[idx], idx, :] = np.tile(labels[idx], (lengths[idx], 1))
        x[idx, :lengths[idx], :] = s  # Fills the matrix with original data,
        #  and the rest remains zero
        x_mask[idx, :lengths[
            idx]] = 1.  # Creates the mask, which is simply ones at all the entries that
        # contains real data

        # Can create a mask here that accommodates for dropouts in the data
        # drop_mask[np.nonzero(s)[0], idx, np.nonzero(s)[1]] = 1. #I don't think the padded sequence
        # steps will influence calculations, because the values will be made zero with the original
        # mask in any case.
        # drop_mask[lengths[idx]:,idx,:]=1.

    return x, x_mask, np.array(lengths)


def add_gauss_noise(X, percentage=0.2):
    '''Add gaussian noise to a percentage of each dimension in the data

    Input:
    X - Numpy array [n_samples, n_steps, n_inputs]
    percentage - Percentage of random n_steps to add noise to.

    Returns:
    X_noise - Numpy array [n_samples, n_steps, n_inputs] where the percentage of n_steps have added gaussian noise
    '''
    X_noise = X.copy()

    n_samples = X.shape[0]
    n_steps = X.shape[1]

    for dimension in range(X.shape[2]):
        for i in range(n_samples):
            mask = np.random.randint(
                0, n_steps, int(round(percentage * n_steps)))
            for m in mask:
                # Add gaussian noise with zero mean and unit variance to the
                # data.
                X_noise[i, m, dimension] = X[
                    i, m, dimension] + np.random.normal(0, 0.1)

    return X_noise


try:
    import matplotlib.pyplot as plt

    def plot_confusion_matrix(y_true, y_pred, class_names=None,
                              normalize=True,
                              title='Confusion matrix',
                              cmap=plt.cm.Blues,
                              deci=2):
        """
        This function prints and plots a confusion matrix of y and y_hat.
        Normalization can be applied by setting `normalize=True`.

        Input:
        y_true -     The true labels of the data.
        y_pred - The model-predicted labels of the data.
        class_names - A list of the class names that are to be displayed
                      on both axes.
        deci - (int) The number of decimals to display in the plot.
        """

        if len(y_true.shape) == 2:
            y_true = np.argmax(y_true, axis=1)

        if len(y_pred.shape) == 2:
            y_pred = np.argmax(y_pred, axis=1)

        if class_names is None:
            class_names = [str(i) for i in range(np.max(y_true)+1)]

        cnf_matrix = metrics.confusion_matrix(y_true, y_pred)

        if normalize:
            cnf_matrix = cnf_matrix.astype('float') / \
                cnf_matrix.sum(axis=1)[:, np.newaxis]
            print("Normalized confusion matrix")
        else:
            print('Confusion matrix, without normalization')

        thresh = cnf_matrix.max() / 2.

        plt.imshow(cnf_matrix, interpolation='nearest', cmap=cmap)
        plt.title(title)
        plt.colorbar()
        tick_marks = np.arange(len(class_names))
        plt.xticks(tick_marks, class_names, rotation=45)
        plt.yticks(tick_marks, class_names)

        for i, j in itertools.product(range(cnf_matrix.shape[0]),
                                      range(cnf_matrix.shape[1])):
            plt.text(j, i, np.around(cnf_matrix[i, j],deci),
                     horizontalalignment="center",
                     color="white" if cnf_matrix[i, j] > thresh else "black")

        plt.tight_layout()
        plt.ylabel('True label')
        plt.xlabel('Predicted label')


except ImportError:
    print("Could not import matplotlib.pyplot")
