'''
A set of useful operations that require tensorflow
'''
import os
# Turn off the tensorflow warnings about code not being compiled with some
# optimizations. (switch off warning logging, but not error logging)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' # noqa
import tensorflow as tf
import numpy as np


def get_blank_state(batch_size, hidden_dim):
    '''
    A function to generate the initial state for an rnn.

    Input:
    batch_size - (int) the batch size
    hidden_dim - A list of ints, each int in the list represents the number on
    that layer.

    Returns:
    A list of LSTMStateTuples.

    Example usage:
    input_feed = {state_c: tuple(s.c for s in get_blank_state(b, h_dim)),
                  state_h: tuple(s.h for s in get_blank_state(b, h_dim))
                  }
    '''
    blank_state = [tf.contrib.rnn.LSTMStateTuple(
        np.zeros([batch_size, h]).astype(np.float32),
        np.zeros([batch_size, h]).astype(np.float32))
                   for h in hidden_dim] # noqa

    return blank_state


def create_scalar_summaries(tags, values):
    '''
    Input:
    tags - the tag names to use in the summary
    values - the values to log

    Returns:
    A tensorflow summary to be used in wrtier.add_summary(summary, steps)
    '''
    summary_value_list = []
    for i in range(len(tags)):
        summary_value_list.append(tf.Summary.Value(tag=tags[i],
                                                   simple_value=values[i]))
    return tf.Summary(value=summary_value_list)


def orthogonal(shape):
    flat_shape = (shape[0], np.prod(shape[1:]))
    a = np.random.normal(0.0, 1.0, flat_shape)
    u, _, v = np.linalg.svd(a, full_matrices=False)
    q = u if u.shape == flat_shape else v
    return q.reshape(shape)


def orthogonal_initializer():
    def _initializer(shape, dtype=tf.float32, partition_info=None):
        return tf.constant(orthogonal(shape), dtype)
    return _initializer


def linear(input_data, output_dim, scope=None, stddev=1.0, init_func='norm'):
    if init_func == 'norm':
        initializer = tf.random_normal_initializer(stddev=stddev)
    elif init_func == 'orth':
        initializer = orthogonal_initializer()
    const = tf.constant_initializer(0.0)
    with tf.variable_scope(scope or 'linear'):
        w = tf.get_variable(
            'w', [input_data.get_shape()[-1], output_dim],
            initializer=initializer)
        b = tf.get_variable('b', [output_dim], initializer=const)
        return tf.matmul(input_data, w) + b


def batch_linear(input_data, output_dim, scope=None, stddev=1.0):
    '''
    Matrix multiplication with the first matrix having 3 dimensions.
    First dimension is batches.

    Inputs:
    input_data - a tensor of the batch of data
                 with shape (batch_size,n_steps,n_inputs)
    output_dim - (int) the number of elements in the last
                 dimension of the output
    '''
    norm = tf.random_normal_initializer(stddev=stddev)
    const = tf.constant_initializer(0.0)
    with tf.variable_scope(scope or 'batch_linear'):
        w = tf.get_variable('w', [input_data.get_shape()[
                            2], output_dim], initializer=norm)
        b = tf.get_variable('b', [output_dim], initializer=const)
        # Replicate the weight matrix to be in batches similar to input_data
        # Use tf.shape for run-time determination of tensor shape instead
        # of using tensor.get_shape() which is used to get the static
        # dimension.
        w = tf.tile(tf.expand_dims(w, 0), tf.stack(
            [tf.shape(input_data)[0], 1, 1]))

        return tf.matmul(input_data, w) + b


def variable_summaries(var, name='summaries'):
    """Attach a lot of summaries to a Tensor (for TensorBoard visualization)."""
    with tf.name_scope(name):
        mean = tf.reduce_mean(var)
        tf.summary.scalar('mean', mean)
        with tf.name_scope('stddev'):
            stddev = tf.sqrt(tf.reduce_mean(tf.square(var - mean)))
        tf.summary.scalar('stddev', stddev)
        tf.summary.scalar('max', tf.reduce_max(var))
        tf.summary.scalar('min', tf.reduce_min(var))
        tf.summary.histogram('histogram', var)
