import tensorflow as tf
from universal_tf_rnn.aux_code.rnn_cells import BNLSTMCell
from universal_tf_rnn.aux_code.tf_ops import linear
from universal_tf_rnn.aux_code.ops import timeit


def rnn(x, h_dim, y_dim, keep_prob, sequence_lengths, batch_norm,
        training, output_format, initial_state, cell_type='lstm'):
    '''
    Inputs:
    x - The input data.
        Tensor shape (batch_size, max_sequence_length, n_features)
    h_dim - A list with the number of neurons in each hidden layer
    keep_prob - The percentage of weights to keep in each iteration
                 (1-drop_prob)

    Returns:
    The non-softmaxed output of the LSTM.
    Tensor shape (batch_size, max_sequence_length, n_features)
    '''

    def single_cell(dim, output_projection=None):
        if cell_type == 'lstm':
            if not batch_norm:

                cell = tf.contrib.rnn.LSTMCell(dim,
                                               forget_bias=1.0,
                                               num_proj=output_projection)
            else:
                print('Using batch norm')
                cell = BNLSTMCell(dim, training)

            drop_cell = tf.contrib.rnn.DropoutWrapper(
                cell,
                input_keep_prob=1,
                output_keep_prob=keep_prob,
                seed=1)
            return drop_cell

    if output_format == 'last':
        # Save some memory and compuation by
        # not requiring the outputs at each step
        if len(h_dim) > 1:
            cell = tf.contrib.rnn.MultiRNNCell(
                [single_cell(dim) for dim in h_dim])

            _, final_state = tf.nn.dynamic_rnn(
                cell, x, sequence_length=sequence_lengths,
                initial_state=initial_state,
                dtype=tf.float32)
        else:
            cell = single_cell(h_dim[0])

            _, final_state = tf.nn.dynamic_rnn(
                cell, x, sequence_length=sequence_lengths,
                initial_state=initial_state,
                dtype=tf.float32)

        return final_state
    else:
        if len(h_dim) > 1:
            cell = tf.contrib.rnn.MultiRNNCell(
                [single_cell(dim) for dim in h_dim[:-1]] +
                [single_cell(h_dim[-1], y_dim)])
        else:
            cell = single_cell(h_dim[0], y_dim)

        outputs, final_state = tf.nn.dynamic_rnn(
            cell, x, sequence_length=sequence_lengths,
            initial_state=initial_state,
            dtype=tf.float32)

        return outputs, final_state


class RNN_Model(object):

    def __init__(self, n_features, n_classes, h_dim, max_sequence_length,
                 is_test, batch_norm, class_weights=None,
                 max_gradient_norm=None, opt_method='adam',
                 learning_rate=0.001, weight_decay=0,
                 old_model=False):
        self.n_features = n_features
        self.n_classes = n_classes
        self.h_dim = h_dim
        self.max_sequence_length = max_sequence_length

        self.opt_method = opt_method
        if old_model:
            self.learning_rate = learning_rate
        else:
            self.learning_rate = tf.Variable(float(learning_rate),
                                             trainable=False,
                                             dtype=tf.float32)
        self.old_model = old_model

        self.max_gradient_norm = max_gradient_norm
        self.is_test = is_test
        self.batch_norm = batch_norm
        self.class_weights = class_weights
        self.weight_decay = weight_decay

        print("Building model ...")
        self.build()

    def build_inputs(self):
        self.keep_prob = tf.placeholder(tf.float32, name='keep_prob')
        self.x = tf.placeholder(tf.float32, [None,
                                             None,
                                             self.n_features],
                                name='x')
        self.y = tf.placeholder(tf.float32, [None, self.n_classes],
                                name='y')
        self.sequence_lengths = tf.placeholder(tf.int32, [None],
                                               name="sequence_lengths")
        self.training = tf.placeholder(tf.bool)

        self.state_c = tuple(tf.placeholder(tf.float32, [None, self.h_dim[i]],
                                            name='state_c' + str(i))
                             for i in range(len(self.h_dim)))
        self.state_h = tuple(tf.placeholder(tf.float32, [None, self.h_dim[i]],
                                            name='state_h' + str(i))
                             for i in range(len(self.h_dim)))
        self.states = tuple(tf.contrib.rnn.LSTMStateTuple(self.state_c[i],
                                                          self.state_h[i])
                            for i in range(len(self.h_dim)))

        # If there is only one layer, get the inner tuple
        if len(self.states) == 1:
            self.states = self.states[0]

    def build_loss(self, outputs):
        sample_cross_entropy = tf.nn.softmax_cross_entropy_with_logits(
            labels=self.y, logits=outputs)

        if self.class_weights is not None:
            print('Scaling loss by class weights')
            class_weights = tf.convert_to_tensor(self.class_weights,
                                                 dtype=tf.float32,
                                                 name='class_weights')
            weight_map = tf.multiply(tf.one_hot(tf.nn.top_k(self.y).indices,
                                                depth=self.n_classes,
                                                dtype=tf.float32),
                                     class_weights)
            weight_map = tf.reduce_sum(weight_map, axis=-1)
            sample_cross_entropy = tf.multiply(sample_cross_entropy,
                                               weight_map)

        weight_decay = self.weight_decay * tf.add_n([tf.nn.l2_loss(v) for v in
                                                     tf.trainable_variables()])
        tf.summary.scalar('weight_decay', weight_decay)
        tf.summary.scalar('cross_entropy',
                          tf.reduce_mean(sample_cross_entropy))

        self.loss = tf.reduce_mean(sample_cross_entropy) + weight_decay

    def build_optimizer(self):
        if not self.old_model:
            self.learning_rate_decay_op = self.learning_rate.assign(
                self.learning_rate * 0.1)

        if self.opt_method == 'adam':
            print('Optimizing with Adam')
            opt = tf.train.AdamOptimizer(self.learning_rate)
        elif self.opt_method == 'momentum':
            print('Optimizing with momentum')
            opt = tf.train.MomentumOptimizer(self.learning_rate,
                                             momentum=0.9,
                                             use_nesterov=True)

        params = tf.trainable_variables()
        gradients = tf.gradients(self.loss, params)
        clipped_gradients, norm = tf.clip_by_global_norm(
            gradients, self.max_gradient_norm)

        tf.summary.scalar('gradients_norm', norm)

        self.train_opt = opt.apply_gradients(
            zip(clipped_gradients, params))

    def build(self, output_format='last'):
        self.build_inputs()

        if output_format == 'last':
            self.final_state = rnn(self.x, self.h_dim,
                                   self.n_classes, self.keep_prob,
                                   sequence_lengths=self.sequence_lengths,
                                   batch_norm=self.batch_norm,
                                   training=self.training,
                                   output_format=output_format,
                                   initial_state=self.states)
            # self.final_state is a tuple with
            # (state.c, state.h)
            # We only use state.c to compute the outputs
            if len(self.h_dim) > 1:
                # If we have a multi-layer rnn, get the top layer state
                top_c = self.final_state[-1][0]
                # Proved: the final entry in the tuple of states returned is the
                # top layer state of the RNN.
            else:
                top_c = self.final_state[0]
            outputs = linear(top_c, self.n_classes,
                             init_func='orth')

        elif output_format == 'average':
            outputs, self.final_state = rnn(
                self.x, self.h_dim, self.n_classes, self.keep_prob,
                sequence_lengths=self.sequence_lengths,
                training=self.training, initial_state=tf.tuple(self.states),
                output_format=output_format)
            # outputs have shape (bath_size, max_sequence_length, n_classes)

            outputs = tf.reduce_mean(outputs, axis=1)

        self.build_loss(outputs)

        self.output_probs = tf.nn.softmax(outputs)

        if not self.is_test:
            print("Adding training operations")
            self.build_optimizer()

        self.summary_op = tf.summary.merge_all()

    @timeit
    def train_step(self, sess, x, y, seq_lengths, keep_prob):
        input_feed = {self.x: x,
                      self.y: y,
                      self.sequence_lengths: seq_lengths,
                      self.training: True,
                      self.keep_prob: keep_prob}

        outputs = sess.run([self.train_opt,
                            self.loss,
                            self.output_probs],
                           input_feed)
        return outputs[-2:]
